const State = @import("./State.zig");

const coral = @import("coral");

state: *State,
fields: FieldTable,
array: ArrayList,

const ArrayList = coral.list.Stack(State.Variant);

const Field = struct {
	field: *State.Object,
	value: State.Variant,

	fn release_objects(self: Field, state: *State) void {
		state.release(self.field);

		if (self.value == .object) {
			state.release(self.value.object);
		}
	}
};

const FieldTable = coral.map.Table([]const coral.io.Byte, Field, coral.map.string_table_traits);

const Self = @This();

pub fn free(self: *Self) void {
	{
		var field_iterator = FieldTable.Iterable{.table = &self.fields};

		while (field_iterator.next()) |entry| {
			entry.value.release_objects(self.state);
		}
	}

	self.fields.free();
	self.array.free();
}

pub fn get_field(self: *Self, field_name: *State.Object) State.Variant {
	const field = self.fields.lookup(field_name.userdata) orelse return .nil;

	if (field.value == .object) {
		return .{.object = self.state.acquire_instance(field.value.object)};
	}

	return field.value;
}

pub fn get_index(self: *Self, index: usize) State.Variant {
	return self.array.values[index];
}

pub fn make(allocator: coral.io.Allocator, state: *State) Self {
	return .{
		.state = state,
		.fields = FieldTable.make(allocator),
		.array = ArrayList.make(allocator),
	};
}

pub fn set_field(self: *Self, field_name: *State.Object, value: State.Variant) coral.io.AllocationError!void {
	const previous_entry = try self.fields.replace(field_name.userdata, .{
		.field = field_name,
		.value = value,
	});

	if (previous_entry) |entry| {
		entry.value.release_objects(self.state);
	}
}

pub fn set_index(self: *Self, index: usize, value: State.Variant) coral.io.AllocationError!void {
	self.array.values[index] = value;
}
